library(ini)


convert_value <- function(value, field_info) {

    if (is.null(field_info) || nrow(field_info) == 0) {
        return (value)
    }

    if (field_info$type == "numeric" || field_info$type == "integer") {
        value <- as.numeric(value)

        return (value)
    } else if (field_info$type == "logical") {
        if (startsWith(value, "y") || startsWith(value, "Y") ||
            startsWith(value, "1") ||
            startsWith(value, "t") || startsWith(value, "T")
        ) {
            return(TRUE)
        } else {
            return(FALSE)
        }
    } else if (field_info$type == "list") {
        if (!is.character(value)) {
            warning(paste("Validation failed for is.character(): '", base::toString(value), "', attempting to fix", sep=""))
            value <- as.character(value)

        }

        value <- unlist(strsplit(value, ",", fixed = T))

        if (field_info$subtype == "numeric" || field_info$subtype == "integer") {
            value <- list(unlist(lapply(value, FUN = as.numeric)))
        } else {
            value <- list(value)
        }


        return(value)
    } else {
        return(value)
    }
    return(value)
}


get_internal_section_name <- function(section_name, identifier) {
    r_section_name <- section_name

    if (section_name == "environment") {
        r_section_name <- "suitability"
    }

    if (section_name == "overlay") {
        r_section_name <- sprintf("overlay_%s", trimws(identifier))
    }

    return(r_section_name)
}


#' Load a CATS configuration file
#'
#' Loads a CATS configuration file (in INI formats) and returns a
#' CATS configuration data frame.
#'
#' @param file.name The path of the INI file to load
#'
#' @return A data frame containing the configuration.
#' @export
#'
load_cats_configuration <- function(file.name) {
    sections <- data.frame(name = c(), identifier = c(), full = c())
    data <- ini::read.ini(file.name)

    for (section in names(data)) {
        items <- unlist(strsplit(section, ":", fixed = T))
        if (length(items) > 2) {
            stop(sprintf("Invalid file format, to many colons in %s", section))
        }
        if (length(items) == 2) {
            name <- items[1]
            identifier <- items[2]
            full <- section
        } else {
            name <- items[1]
            identifier <- NA
            full <- section
        }

        sections <- rbind(sections, data.frame(name = trimws(name), identifier = trimws(identifier), full = full))
    }

    fields <- cats_config_fields

    species_name <- as.character(sections[which(sections$name == "species"), "identifier"])


    ignored_keys <- c("species.suitability", "species.dispersal")
    config <- cats_default_config(species.name = species_name)
    hybrid_mode <- F

    for (row in 1:nrow(sections)) {
        section_full <- sections[row, "full"]
        section_name <- sections[row, "name"]
        identifier <- sections[row, "identifier"]
        r_section_name <- get_internal_section_name(section_name, identifier)



        keys <- names(data[[section_full]])

        items <- data[[section_full]]


        for (key in keys) {
            r_key <- paste(r_section_name, gsub(" ", "_", key, fixed = T), sep = ".")
            value <- as.character(items[[key]])

            if (r_key == "suitability.type" && value == "suitability") {
                hybrid_mode <- T
                next
            }

            if (r_key %in% ignored_keys) {
                next
            }


            field_info <- fields[which(fields$field == r_key), ]

            if (nrow(field_info) == 0) {
                config[[r_key]] <- value
                next
            }

            value < convert_value(value, field_info)
            config[[r_key]] <- value
        }
    }

    if (!hybrid_mode) {
        stop("Could not load configuration file, currently only hybrid mode is supported")
    }
    return(config)
}
