get_run_name <- function(catsconfig)
{
    cc <- NA
    config <- catsconfig

    if (class(catsconfig) == "character") {
        if (file.exists(catsconfig)) {
            config <- load_cats_configuration(catsconfig)
        } else {
            warning(sprintf("Could not load CATS configuration file %s", catsconfig))
        }
    }

    if (class(config) == "data.frame") {
        run_name <- config$general.run_name
}

    if (is.null(run_name)  || is.na(run_name))
    {
        warning("Could not run name from configuration")
    }

    return(run_name)

}

get_cats_binary <- function(path_to_cats) {
    path_to_cats <- normalizePath(path_to_cats)

    bin_name <- "cats"
    windows <- F

    if (.Platform$OS.type == "windows") {
        windows <- T
        bin_name <- "cats.exe"
    }


    if (R.utils::isDirectory(path_to_cats)) {
        cats <- file.path(path_to_cats, bin_name)
    } else {
        cats <- path_to_cats
    }

    if (!file.exists(cats)) {
        if (R.utils::isDirectory(path_to_cats)) {
            cats <- file.path(path_to_cats, "cats-windows", "cats.exe")
        }
    }

    if (!file.exists(cats)) {
        stop(sprintf("Could not find cats executable in %s", path_to_cats))
    }

    cats <- normalizePath(cats)
    return(cats)
}


#' Run CATS simulations
#'
#' @details
#' The replicate number(s) are used to differentiate multiple CATS runs with the same configuration.
#' The output of these is usually not the same due to stochastic effects.
#' If `replicate_numbers` is a vector, then one simulation is run for each element in this vector.
#'
#' @param configuration either a file name of a CATS configuration file or
#' a CATS configuration data frame
#' @param working_directory path to the working directory for the CATS simulation to be run in.
#' All relative paths in the CATS configuration are relative to this directory.
#' @param path_to_cats path to the CATS binary
#' @param replicate_numbers a vector or a single replicate number, i.e. integers greater or equal to zero
#' Default: 0
#' @param quiet logical, default TRUE. If set to FALSE, the output of the CATS simulation will
#' be printed in the CATS session.
#'
#' @return A data frame containing the output files of a all started CATS runs,
#' which can be plotted with [plot_cats_run()]
#' Each row of the data frame corresponds to a single output file. The columns
#' of the data frame are the run name, the replicate number, the year (or NA if not applicable, e.g. for log files),
#' the type of output file("log", "grid-stats", "population") and the file name.
#' 'log' is the command line output of the CATS simulation,
#' 'grid-stats' is a CSV file with summary statistics for a CATS simulation, and
#' 'population' are population abundance files in GeoTIFF format.
#' @export
run_cats <- function(configuration, working_directory, path_to_cats, replicate_numbers = c(0), quiet = F) {
    cats <- get_cats_binary(path_to_cats)

    result <- NULL

    if (class(replicate_numbers) != "numeric") {
        stop("replicate_numbers must be numeric")
    }

    for (x in replicate_numbers) {
        if (x < 0 || as.integer(x) != as.numeric(x) ) {
            stop("replicate_numbers must be integer-valued and >= 0")
        }
    }


    if (class(configuration) == "data.frame") {
        check_cats_configuration(configuration)

        for (row in 1:nrow(configuration)) {
            conf <- configuration[row, ]
            conf_file <- write_cats_config(conf, working_directory = working_directory)
            run_name <- conf$general.run_name

            for (replicate in replicate_numbers) {
                result <- run_cats_internal(result, working_directory, run_name, cats, conf_file, replicate = replicate, quiet)
            }
        }
    } else if (class(configuration) == "character") {
        config <- configuration
        if (!file.exists(configuration)) {
            config <- file.path(working_directory, configuration)

        }
        config <- normalizePath(config)
        if (!file.exists(config)) {
            stop(sprintf("Could not find configuration file %s", config))
        }
        run_name <- get_run_name(config)
        for (replicate in replicate_numbers) {

            result <- run_cats_internal(result, working_directory, run_name, cats, config, replicate = replicate, quiet)
        }
    } else {
        stop("Don't know what to do with the supplied configuration data")
    }


    return(result)
}


run_cats_internal <- function(result, working_directory, run_name, cats_binary, conf_file, replicate, quiet) {
    working_directory <- normalizePath(working_directory)
    summary_file <- normalizePath(file.path(working_directory, sprintf("catsr-summary-%s-%03d.csv", make_filename_valid(run_name), replicate)), mustWork = F)

    log_file <- normalizePath(file.path(working_directory, sprintf("catsr-log-%s-%03d.txt", make_filename_valid(run_name), replicate)), mustWork = F)
    conf_file <- normalizePath(conf_file)

    args <- c(conf_file, "--summary-file", summary_file, "--log-file", log_file, "--replicate", replicate)

    if (quiet) {
        args <- c(args, "--quiet")
    }

    saved_wd <- getwd()

    result <- tryCatch({
        cat("Running CATS\n")
        cat(sprintf(" -- working directory '%s'\n -- command line: \"cats %s\"\n", working_directory, paste(args, sep = " ", collapse = " ")))
        cat(sprintf(" -- log file will be written to %s\n", log_file))
        cat(sprintf(" -- list of output files will be written to %s\n", summary_file))
        setwd(normalizePath(working_directory))
        cat(sprintf(" -- simulation started\n"))
        system2(cats_binary, args = args)
        cat("\n\n")


        output <- utils::read.csv(summary_file, sep = ",", header = T)
        for (r in 1:nrow(output)) {
            fn <- output[r, "file.name"]
            if (!R.utils::isAbsolutePath(fn)) {
                fn <- normalizePath(file.path(working_directory, fn))
                output[r, "file.name"] <- fn
            }
        }

        if (is.null(result)) {
            result <- output
        } else {
            result <- rbind(result, output)
        }
    }, finally = {
        setwd(saved_wd)
    })
    return(result)
}
