valid_excerpt <- function(excerpt) {
    if (class(excerpt) != "data.frame") {
        stop("Expected data.frame")
    }
    if (nrow(excerpt) != 1) {
        stop("Expected data.frame with a single row")
    }
}


verify_suitability <- function(excerpt, working_directory) {
    valid_excerpt(excerpt)
    result <- cats_validation_object()
    if (is.null(excerpt$suitability.filename_pattern)) {
        return(result)
    }
    if (excerpt$suitability.interpolation_interval > 0 && !grepl("%Y", excerpt$suitability.filename_pattern, fixed = T)) {
        error <- c("suitability.interpolation_interval greater than 0 requires that config.suitability.filename_pattern contains '%Y', which gets substituted with current simulation year")
        result <- add_error(result, error)
    }
    if (endsWith(excerpt$suitability.filename_pattern, ".tif") || endsWith(excerpt$suitability.filename_pattern, ".tiff")) {

    } else {
        result <- add_warning(result, sprintf("suitability.filename_pattern '%s' should be in GeoTIFF format (.tif or .tiff)", excerpt$suitability.filename_pattern))
    }

    return(result)
}


verifiy_inital_population_filename <- function(excerpt, working_directory) {
    valid_excerpt(excerpt)
    result <- cats_validation_object()
    if (is.null(excerpt$species.initial_population_filename)) {
        return(result)
    }

    if (endsWith(excerpt$species.initial_population_filename, ".tif") || endsWith(excerpt$species.initial_population_filename, ".tif")) {

    } else {
        result <- add_warning(result, sprintf("species.initial_population_filename should be in GeoTIFF format (.tif or .tiff)", excerpt$species.initial_population_filename))
    }

    return(result)
}

cats_validation_object <- function(catsconfig) {
    errors <- 0
    warnings <- 0

    return(list(errors = errors, warnings = warnings, error_messages = c(), warning_messages = c()))
}

add_warning <- function(validation, message) {
    validation$warning_messages <- c(validation$warning_messages, message)
    validation$warnings <- validation$warnings + 1
    return(validation)
}

add_error <- function(validation, message) {
    validation$error_messages <- c(validation$error_messages, message)
    validation$errors <- validation$errors + 1
    return(validation)
}

add_validation <- function(v1, v2) {
    v <- cats_validation_object()
    v$errors <- v1$errors + v2$errors
    v$warnings <- v1$warnings + v2$warnings
    v$error_messages <- c(v1$error_messages, v2$error_messages)
    v$warning_messages <- c(v1$warning_messages, v2$warning_messages)
    return(v)
}



check_dispersal <- function(catsconfig) {
    valid_excerpt(catsconfig)
    valid <- cats_validation_object()

    minima <- catsconfig[, "dispersal.minimum_weights"]
    maxima <- catsconfig[, "dispersal.maximum_weights"]

    minima_info <- cats_config_fields[which(cats_config_fields$field == "dispersal.minimum_weights"), ]
    maxima_info <- cats_config_fields[which(cats_config_fields$field == "dispersal.maximum_weights"), ]

    minima <- convert_value(minima, minima_info)
    maxima <- convert_value(maxima, minima_info)

    if (class(minima) == "list") {
        minima <- minima[[1]]
    }
    if (class(maxima) == "list") {
        maxima <- maxima[[1]]
    }



    if (length(minima) != length(maxima)) {
        valid <- add_error(valid, "dispersal.minimum_weights and dispersal.maximum_weights must have the same lengths")
        return(valid)
    }
    min_sum <- 0
    max_sum <- 0

    for (i in 1:length(minima)) {
        # maxima have already been checked using the standard function
        if (is.na(minima[i])) {
            valid <- add_error(valid, "dispersal.minimum_weights not specified")
            break
        }

        if (is.na(maxima[i])) {
            valid <- add_error(valid, "dispersal.maximum_weights not specified")
            break
        }
        if (minima[i] < 0 && minima[i] != -1) {
            valid <- add_error(valid, "all entries of dispersal.minimum_weights must be greater or equal than 0 (-1 is allowed to indicate the remaining weight, only valid for the last entry)")
        }

        if (maxima[i] < 0 && maxima[i] != -1) {
            valid <- add_error(valid, "all entries of dispersal.maximum_weights must be greater or equal than 0 (-1 is allowed to indicate the remaining weight, only valid for the last entry)")
        }

        if ((minima[i] == -1 || maxima[i] == -1) && i != length(minima)) {
            valid <- add_error(valid, "only the last entry of dispersal.minimum_weights and dispersal.maximum_weights is allowed to be -1")
        }

        if (maxima[i] < minima[i] && minima[i] != -1 && maxima[i] != -1) {
            valid <- add_error(valid, "all elements of dispersal.minimum_weights must be less or equal than the corresponding elements in dispersal.maximum_weights")
        }

        if (minima[i] != -1) {
            min_sum <- min_sum + minima[i]
        }
        if (maxima[i] != -1) {
            max_sum <- max_sum + maxima[i]
        }
    }
    if (min_sum > 1) {
        valid <- add_error(valid, sprintf("dispersal.minimum_weights add up to more than 1.0: %f", min_sum))
    }
    if (max_sum > 1) {
        valid <- add_error(valid, sprintf("dispersal.maximum_weights add up to more than 1.0: %f", max_sum))
    }




    return(valid)
}

check_configuration_fields <- function(catsconfig, field, value) {
    valid_excerpt(catsconfig)
    valid <- cats_validation_object()

    allowed <- c()
    fields <- get_cats_hybrid_fields()
    cats_mode <- "hybrid"
    if (!is.null(attributes(catsconfig)$comment)) {
        cats_mode <- attributes(catsconfig)$comment
    }

    if (cats_mode == "hybrid") {
        mode <- fields$mode == "hybrid"
        required <- fields$required == T
        allowed <- fields[mode, ]$field
        req <- fields[mode & required, ]$field
    }



    if (field %in% allowed) {
    } else {
        valid <- add_warning(valid, sprintf("unknown field %s, will most likely be ignored", field))
    }


    if (field %in% colnames(catsconfig)) {

    } else {
        valid <- add_error(valid, sprintf("required field '%s' not found", field))
    }


    value <- catsconfig[, field]
    minimum <- fields[field == fields$field, "minimum"]
    maximum <- fields[field == fields$field, "maximum"]

    if (class(value) == "list") {
        for (val in value[[1]]) {
            if (is.na(minimum) == FALSE && length(minimum) > 0 && val < minimum) {
                valid <- add_error(valid, sprintf("value of field '%s' is %f < allowed minimum of %f", field, val, minimum))
            }

            if (is.na(maximum) == FALSE && length(maximum) > 0 && val > maximum) {
                valid <- add_error(valid, sprintf("value of field '%s' is %f > allowed maximum of %f", field, val, maximum))
            }
        }
    } else {
        if (!is.na(minimum) && length(minimum) > 0 && (is.na(value) || value < minimum)) {
            valid <- add_error(valid, sprintf("value of field '%s' is %f < allowed minimum of %f", field, value, minimum))
        }

        if (!is.na(maximum) && length(maximum) > 0 && (is.na(value) || value > maximum)) {
            valid <- add_error(valid, sprintf("value of field '%s' is %f > allowed maximum of %f", field, value, maximum))
        }
    }




    return(valid)
}


check_single_cats_configuration <- function(catsconfig) {
    valid_excerpt(catsconfig)
    valid <- cats_validation_object()

    allowed <- c()
    fields <- get_cats_hybrid_fields()
    cats_mode <- "hybrid"
    if (!is.null(attributes(catsconfig)$comment)) {
        cats_mode <- attributes(catsconfig)$comment
    }

    if (cats_mode == "hybrid") {
        mode <- fields$mode == "hybrid"
        required <- fields$required == T
        allowed <- fields[mode, ]$field
        req <- fields[mode & required, ]$field
    }


    for (field in colnames(catsconfig)) {
        if (field %in% allowed) {
        } else {
            valid <- add_warning(valid, sprintf("unknown field %s, will most likely be ignored", field))
        }
    }

    for (field in req) {
        if (field %in% colnames(catsconfig)) {

        } else {
            valid <- add_error(valid, sprintf("required field '%s' not found", field))
        }
    }

    for (field in colnames(catsconfig))
    {
        name <- field
        value <- catsconfig[, field]
        minimum <- fields[field == fields$field, "minimum"]
        maximum <- fields[field == fields$field, "maximum"]
        field_info <- cats_config_fields[which(cats_config_fields$field == name), ]
        if (!is.null(field_info) && nrow(field_info) > 0) {
            value <- convert_value(value, field_info)
        }



        if (class(value) == "list") {
            for (val in value[[1]]) {
                if (is.na(minimum) == FALSE && length(minimum) > 0 && val < minimum) {
                    valid <- add_error(valid, sprintf("value of field '%s' is %f < allowed minimum of %f", name, val, minimum))
                }

                if (is.na(maximum) == FALSE && length(maximum) > 0 && val > maximum) {
                    print(name, val, maximum)
                    valid <- add_error(valid, sprintf("value of field '%s' is %f > allowed maximum of %f", name, val, maximum))
                }
            }
        } else {

            if (!is.na(minimum) && length(minimum) > 0 && (is.na(value) || value < minimum)) {
                valid <- add_error(valid, sprintf("value of field '%s' is %f < allowed minimum of %f", name, value, minimum))
            }

            if (!is.na(maximum) && length(maximum) > 0 && (is.na(value) || value > maximum)) {
                valid <- add_error(valid, sprintf("value of field '%s' is %f > allowed maximum of %f", name, value, maximum))
            }

        }
    }

    valid <- add_validation(valid, check_dispersal(catsconfig))
    valid <- add_validation(valid, check_filenames(catsconfig))


    return(valid)
}


check_validation <- function(valid) {
    if (valid$warnings > 0) {
        print(sprintf("%d warnings found", valid$warnings))

        for (i in 1:length(valid$warning_messages)) {
            print(sprintf("Warning %d of %d: %s", i, valid$warnings, valid$warning_messages[[i]]))
        }
    }

    if (valid$errors > 0) {
        for (i in 1:length(valid$error_messages)) {
            print(sprintf("Error %d of %d: %s", i, valid$errors, valid$error_messages[[i]]))
        }

        stop(sprintf("Stopping: %d errors", valid$errors))
    }
}


check_cats_configuration <- function(catsconfig) {
    valid <- cats_validation_object()


    for (row in nrow(catsconfig)) {
        config <- catsconfig[row, ]
        valid <- add_validation(valid, check_single_cats_configuration(config))
    }
}
